#include "srcguru.h"
#include "addropt.h"
#include "eval.h"
#include "params.h"

/* Scans to determine if register's address is used */

int is_addr_used(source_item *scan, int x)
{
    while (scan) {
        if (scan->type == ITEM_INSTRUCTION) {
            if (scan->addrset[x] == AC_NONE)
                return 0;
            if (scan->addruse[x])
                return 1;
        }
        scan = scan->next;
    }
    return 0;
}

/* Main address optimization - elimantes unnecessary conversion of Z80
   address to 68000 addresses */

void addr_optimize()
{
    int t1, r1, p1, p2, r2, t2, x, y, z;
    int addrset[8];
    char o1[MAX_LINE], o2[MAX_LINE];
    source_item *current;

    /* Determine address register calculation and use */

    for (x = 0; x < 8; addrset[x++] = AC_NONE);

    current = z80_source;
    while (current) {
        current = current->next;

        if (current->type == ITEM_INSTRUCTION) {

            /* First step in instruction handling --> decode params */

            decode_param(a1, &t1, &r1, &p1, o1);
            decode_param(a2, &t2, &r2, &p2, o2);

            /* Set use according to dereferenced registers */

            current->addruse = malloc(sizeof(int) * 8);
            for (x = 0; x < 8; current->addruse[x++] = 0);
            if (t1 == PT_DEREF) 
                current->addruse[r1] = 1;
            if (t2 == PT_DEREF)
                current->addruse[r2] = 1;

            /* Transform any AC_CHECK (beginning of tentaive section) regs
               to AC_TENT (middle of tentative section) regs */

            for (x = 0; x < 8; x++)
                if (addrset[x] == AC_CHECK)
                    addrset[x] = AC_TENT;

            /* If you load a register with a constant, becomes AC_CHECK */

            if (eq(txt, "ld"))
                if (t1 == PT_WORDREG)
                    if (t2 == PT_CONST)
                        addrset[r1] = AC_CHECK;

            /* If you increment or decrement a register, becomes AC_CHECK */

            if ((eq(txt, "inc")) || (eq(txt, "dec")))
                if (t1 == PT_WORDREG) 
                    if (addrset[r1] != AC_NONE)
                        addrset[r1] = AC_CHECK;

            /* Next, save info on available addresses */

            current->addrset = malloc(sizeof(int) * 8);
            for (x = 0; x < 8; x++)
                current->addrset[x] = addrset[x];

            /* Changing the contents of a register another way -> AC_NONE */

            if ((eq(txt, "ldi")) ||
                (eq(txt, "ldir")) ||
                (eq(txt, "ldd")) ||
                (eq(txt, "lddr"))) {
                addrset[2] = AC_NONE;
                addrset[3] = AC_CALC;
                addrset[4] = AC_CALC;
                continue;
            }

            if (eq(txt, "ex")) {
                addrset[3] = AC_NONE;
                addrset[4] = AC_NONE;
                continue;
            }

            if (eq(txt, "add")) 
                if (t1 == PT_WORDREG) {
                    addrset[r1] = AC_NONE;
                    continue;
                }

            if ((eq(txt, "inc")) || (eq(txt, "dec")))
                if (t1 == PT_BYTEREG) {
                    addrset[r1] = AC_NONE;
                    continue;
                }

            if (eq(txt, "pop"))
                if (t1 == PT_WORDREG) {
                    addrset[r1] = AC_NONE;
                    continue;
                }

            if (eq(txt, "ld")) {
                if (t1 == PT_WORDREG) 
                    if (t2 != PT_CONST) {
                        addrset[r1] = AC_NONE;
                        continue;
                    }
                if (t1 == PT_BYTEREG) {
                    addrset[r1] = AC_NONE;
                    continue;
                }
            }

            /* All calls must be assumed to destroy registers */

            if (eq(txt, "call")) {
                for (x = 0; x < 8; addrset[x++] = AC_NONE);    
                continue;
            }

            /* Jumps and returns clear tentative calculations */
                          
            if ((eq(txt, "jp")) || (eq(txt, "jr")) || (eq(txt, "ret"))) {
                for (x = 0; x < 8; x++)
                    if (addrset[x] == AC_TENT)
                        addrset[x] = AC_NONE;
                continue;
            }

            /* Dereferencing a register makes it AC_CALC */

            if (t1 == PT_DEREF) {
                addrset[r1] = AC_CALC;
                continue;
            }

            if (t2 == PT_DEREF) {
                addrset[r2] = AC_CALC;
                continue;
            }

        }

        /* If a label is encountered, all registers become unknown */

        if (current->type == ITEM_LABEL)
            for (x = 0; x < 8; addrset[x++] = AC_NONE);    
    }

    /* Determine whether hypothetical calculations were really needed */

    current = z80_source;
    while (current) {
        if (current->type == ITEM_INSTRUCTION)
            for (x = 0; x < 8; x++)
                if (current->addrset[x] == AC_CHECK)
                    if (is_addr_used(current->next, x)) 
                        current->addrset[x] = AC_SPEC;
        current = current->next;
    }

}

