#include <stdio.h>

#include "srcguru.h"

#include "hash.h"

#include "macros.h"

hash_item** macro_hash;

/***************** Initialize built-in macros ******************************/

void init_macros()
{
    macro_hash = new_hash_table(256);
    new_macro("ROM_CALL(addr)", "call addr");
    new_macro("ROM_CALL_Z(addr)", "call z,addr");
    new_macro("ROM_CALL_NZ(addr)", "call nz,addr");
    new_macro("ROM_CALL_C(addr)", "call c,addr");
    new_macro("ROM_CALL_NC(addr)", "call nc,addr");
    new_macro("CALL_(addr)", "call addr");
    new_macro("CALL_Z(addr)", "call z,addr");
    new_macro("CALL_NZ(addr)", "call nz,addr");
    new_macro("CALL_C(addr)", "call c,addr");
    new_macro("CALL_NC(addr)", "call nc,addr");
    new_macro("JUMP_(addr)", "jp addr");
    new_macro("JUMP_Z(addr)", "jp z,addr");
    new_macro("JUMP_NZ(addr)", "jp nz,addr");
    new_macro("JUMP_C(addr)", "jp c,addr");
    new_macro("JUMP_NC(addr)", "jp nc,addr");  
}

/***************** Add macro if not already present ************************/

void add_macros(int type, char* name, char* value)
{
    if (get_macro(name) == NULL)
        hash_add(type, name, value, NULL, macro_hash);
}

/***************** Retrieve a macro ****************************************/

hash_item* get_macro(char* name)
{
    return hash_get(name, macro_hash);
}

/***************** Find the end of something in the macro params list ******/

char* end_scan(char* data)
{   
    if (data == NULL)
        return NULL;
    while (*data) {
        if (*data == '"')
            while ((*data) && (*data != '"'))
                data++;
        if (*data == ',')
            break;
        if (*data == ')')
            break;
        data++;
    }
    return data;
}

/***************** Copy the name only of a macro call (returning *params) **/

char* macro_name(char* in, char* out, int* paren)
{
    while ((*in) && (*in != '('))
       *out++ = *in++;
    *paren = (*in == '(');
    *out = 0;
    return in;
}

/***************** Insert a new macro **************************************/

void new_macro(char* def, char* val)
{
    char val2[MAX_LINE];
    int x = 0;
    int y;
    unsigned char* out;
    char* scan;
    char* pstart[16];

    scan = macro_name(def, def, &y) + 1;

    if (!(y)) {
        add_macros(0, def, val);
        return;
    }

    if (pstart != NULL) {              /* pstart - pointers to param names */
        pstart[x++] = scan;
        while (*scan) {
            if (*scan == ',') {
                *scan = 0;
                pstart[x++] = scan + 1;
            }
            if (*scan == ')') {
                *scan = 0;
                break;                    
            }
            scan++;
        }
    }
            
    scan = val;                         /* replace param n with char 240+n */
    out = val2;
    while (*scan) {
        for (y = 0; y < x; y++) 
            if (*pstart[y])
                if (!(strnicmp(pstart[y], scan, strlen(pstart[y])))) {
                    *out++ = 240 + y;
                    scan += strlen(pstart[y]);
                    break;
                }
        if (y == x)
            *out++ = *scan++;
    }
    *out = 0;

    add_macros(x, def, val2);
}

/***************** Check if string is a macro call *************************/

int is_macro(char* str)
{
    char tstr[MAX_LINE];
    hash_item* m;
    int never;

    macro_name(str, tstr, &never);
    m = get_macro(tstr);
    return (m != NULL);    
}

/***************** Retrieve item from macro parameters by number ***********/

void get_item(char* data, int item, char* out)
{
    char* end;
    while (item--) {
        data = end_scan(data + 1);
        if (*data == 0) {
            *out = 0;
            return;
        }
    }     
    end = end_scan(data + 1);
    data++;
    while (data != end)
        *out++ = *data++;
    *out = 0;
}

/***************** Decode a macro call *************************************/

void decode_macro(char* in, char* out)
{
    char item[MAX_LINE];
    char* itemp;
    hash_item* m;
    unsigned char* me;
    char* w;
    int never;

    w = out;

    in = macro_name(in, out, &never);
    m = get_macro(out);
    me = m->value;

    while (*me) {
        if (*me < 240) {
            *out++ = *me;
        } else {
            get_item(in, *me - 240, item);
            itemp = item;
            while (*itemp)
                *out++ = *itemp++;
        }
        me++;
    }

    *out = 0;
}
