#include "srcguru.h"
#include "hash.h"
#include "eval.h"
#include "params.h"
#include "symbols.h"
#include "addropt.h"

#define VFLAGSAVE if (current->do_save_flags) fprintf(outfile, "        move.w  sr,(a0)\n");
#define LEA(reg) if (current->addrset[reg] == AC_NONE) fprintf(outfile, "        __ADDR  %d\n", reg); else fprintf(outfile,"        ;__ADDR %d\n", reg);

/****************** Main instruction converter **************************/

int convert_instruction(source_item* current)
{
    int ptype1, reg1, part1, ptype2, reg2, part2, x, y, z;
    char addr[MAX_LINE], offset1[MAX_LINE], offset2[MAX_LINE], where[25], cmd[15], cmd2[15], dest[MAX_LINE];
    char* i;
    static int ncc = 0;
    ev* expr;

    /************** RLCA, RRCA, RLA, RRA */

    if ((eq(txt, "rlca")) || (eq(txt, "rrca")) || (eq(txt, "rla")) || (eq(txt, "rra"))) {
        txt[strlen(txt) - 1] = 0;
        strcpy(a1, "a");
    }

    /************** EI, DI */

    if ((eq(txt, "ei")) || (eq(txt, "di")))
        return 0;

    /************** HALT */

    if (eq(txt, "halt")) {
        if (turbo)
            fprintf(outfile, "        jsr     ti85lib3@0001\n");
        else
            fprintf(outfile, "        jsr     ti85lib2@0002\n");
        return 0;
    }


    /************** SCF */

    if (eq(txt, "scf")) {
        fprintf(outfile, "        or.w    #1,(a0)\n");
        return 0;
    }

    /************** CCF */

    if (eq(txt, "ccf")) {
        fprintf(outfile, "        eor.w   #1,(a0)\n");
        return 0;
    }

    /************** EX DE, HL */

    if (eq(txt, "ex"))
        if (eq(a1, "de"))
            if (eq(a2, "hl")) {
                fprintf(outfile, "        exg     d3,d4\n");
                return 0;
            }

    /************** RET */

    if (eq(txt, "ret")) {
        strcpy(cmd2, empty);
        decode_cond(a1, cmd2);

        if (*cmd2) {
            if (!current->gets_flags)   
                fprintf(outfile, "        move.b  (a0),ccr\n");         

            fprintf(outfile, "        %s     NC__%d\n", cmd2, ncc);
            if (exact_stack)
                fprintf(outfile, "        __RTS\n");
            else
                fprintf(outfile, "        rts\n");
            fprintf(outfile, "NC__%d:\n", ncc++);
        } else 
            if (exact_stack)
                fprintf(outfile, "        __RTS\n");
            else
                fprintf(outfile, "        rts\n");
        return 0;
    }
           
    /************** NOP */

    if (eq(txt, "nop")) {
        fprintf(outfile, "        nop\n");
        VFLAGSAVE
        return 0;
    }

    /************** NEG */

    if (eq(txt, "neg")) {
        fprintf(outfile, "        neg.b   d7\n");
        VFLAGSAVE
        return 0;
    }

    /************** CPL */

    if (eq(txt, "cpl")) {
        fprintf(outfile, "        not.b   d7\n");
        VFLAGSAVE
        return 0;
    }

    /************** LDDR */

    if (eq(txt, "lddr")) {
        fprintf(outfile, "        __LDDR\n");
        return 0;
    }

    /************** LDIR */

    if (eq(txt, "ldir")) {
        fprintf(outfile, "        __LDIR\n");
        return 0;
    }

    /************** DECODE ARGUMENTS */

    decode_param(a1, &ptype1, &reg1, &part1, offset1);
    decode_param(a2, &ptype2, &reg2, &part2, offset2);

    /************** IN */

    if (eq(txt, "in")) {
        if (eq(a1, "a"))
            if (ptype2 == PT_MEM) {
                fprintf(outfile, "        move.b  #%s,d1\n", offset2);
                if (turbo)
                    fprintf(outfile, "        jsr     ti85lib3@0008\n");
                else
                    fprintf(outfile, "        jsr     ti85lib2@0008\n");
                return 0;
            }
        if (eq(offset2, "c"))
            if (ptype1 == PT_BYTEREG) {
                fprintf(outfile, "        move.w  d7,-(sp)\n");
                fprintf(outfile, "        move.b  d2,d1\n");
                if (turbo)
                    fprintf(outfile, "        jsr     ti85lib3@0008\n");
                else
                    fprintf(outfile, "        jsr     ti85lib2@0008\n");
                put_bytereg(7, reg1, part1);
                fprintf(outfile, "        move.w  (sp)+,d7\n");
                return 0;
            }
    }

    /************** OUT */

    if (eq(txt, "OUT"))
        if (eq(a2, "a"))
            if (ptype1 == PT_MEM) {
                fprintf(outfile, "        move.b  #%s,d1\n", offset1);
                if (turbo)
                    fprintf(outfile, "        jsr     ti85lib3@0007\n");
                else
                    fprintf(outfile, "        jsr     ti85lib2@0007\n");
                return 0;
            }

    /************** PUSH */

    if (eq(txt, "push")) {
        if (ptype1 == PT_WORDREG) {
            fprintf(outfile, "        move.w  d%d,-(sp)\n", reg1);
            return 0;
        }
        if (ptype1 == PT_AF) {
            fprintf(outfile, "        move.b  d7,d1\n");
            fprintf(outfile, "        lsl.w   #8,d1\n");
            fprintf(outfile, "        move.b  1(a0),d1\n");
            fprintf(outfile, "        move.w  d1,-(sp)\n");
            return 0;
        }
    }

    /************** POP */

    if (eq(txt, "pop")) {
        if (ptype1 == PT_WORDREG) {
            fprintf(outfile, "        move.w  (sp)+,d%d\n", reg1);
            return 0;
        }
        if (ptype1 == PT_AF) {
            fprintf(outfile, "        move.w  (sp)+,d1\n");
            fprintf(outfile, "        move.b  d1,1(a0)\n");
            fprintf(outfile, "        lsr.w   #8,d1\n");
            fprintf(outfile, "        move.b  d1,d7\n");
            return 0;
        }
    }

    /************** ADD */

    if (eq(txt, "add")) {
        if (ptype1 == PT_WORDREG) {
            if (ptype2 == PT_WORDREG) {
                fprintf(outfile, "        add.w   d%d,d%d\n", reg2, reg1);
                return 0;
            }
        }
        if (ptype1 == PT_BYTEREG) {
            if (ptype2 == PT_BYTEREG) {
                get_bytereg(reg2, part2, where);
                fprintf(outfile, "        add.b   %s,d7\n", where);
                VFLAGSAVE
                return 0;
            }
            if (ptype2 == PT_CONST) {
                fprintf(outfile, "        add.b   #%s,d7\n", offset2);
                VFLAGSAVE
                return 0;
            }
            if (ptype2 == PT_DEREF) {
                LEA(reg2)
                fprintf(outfile, "        add.b   %s(a%d),d7\n", offset2, reg2);
                VFLAGSAVE
                return 0;
            }
        }
    }

    /************** LD */

    if (eq(txt, "ld")) {
        if (ptype1 == PT_DEREF) {
            LEA(reg1)
            if (ptype2 == PT_BYTEREG) {
                get_bytereg(reg2, part2, where);
                fprintf(outfile, "        move.b  %s,%s(a%d)\n", where, offset1, reg1);
                return 0;
            }
            if (ptype2 == PT_CONST) {
                fprintf(outfile, "        move.b  #%s,%s(a%d)\n", offset2, offset1, reg1);
                return 0;
            }
        }
        if (ptype1 == PT_BYTEREG) {
            if (eq(a2, "r")) {
                fprintf(outfile, "        __RAND\n");
                return 0;
            }
            if (ptype2 == PT_BYTEREG) {
                if ((part1 == RP_HIGH) && (part2 == RP_HIGH)) {
                    fprintf(outfile, "        move.w  d%d,d1\n", reg2);
                    fprintf(outfile, "        and.w   #$ff00,d1\n");
                    fprintf(outfile, "        and.w   #$ff,d%d\n", reg1);
                    fprintf(outfile, "        or.w    d1,d%d\n", reg1);
                }
                if ((part1 == RP_HIGH) && (part2 == RP_LOW)) {
                    fprintf(outfile, "        move.w  d%d,(a1)\n", reg1);
                    fprintf(outfile, "        move.b  d%d,(a1)\n", reg2);
                    fprintf(outfile, "        move.w  (a1),d%d\n", reg1);
                }
                if ((part1 == RP_LOW) && (part2 == RP_HIGH)) {
                    get_bytereg(reg2, part2, where);
                    fprintf(outfile, "        move.b  (a1),d%d\n", reg1);
                }
                if ((part1 == RP_LOW) && (part2 == RP_LOW)) 
                    fprintf(outfile, "        move.b  d%d,d%d\n", reg2, reg1);
                return 0; 
            }
            if (ptype2 == PT_CONST) {
                if (part1 == RP_LOW)
                    if ((eq(offset2, "0")) && (!zkart3dhack)) 
                        fprintf(outfile, "        clr.b   d%d\n", reg1);
                    else
                        fprintf(outfile, "        move.b  #%s,d%d\n", offset2, reg1);
                else {
                    if (eq(offset2, "0"))
                        fprintf(outfile, "        and.w   #$ff,d%d\n", reg1);
                    else {
                        fprintf(outfile, "        and.w   #$ff,d%d\n", reg1);
                        fprintf(outfile, "        or.w    #(%s)<<8,d%d\n", offset2, reg1);
                    }
                }
                return 0;
            }
            if (ptype2 == PT_MEM) {
                load_addr(offset2, 1, addr);
                move_mem(addr, reg1, part1);
                return 0;
            }

            if (ptype2 == PT_DEREF) {
                LEA(reg2);
                sprintf(offset1, "%s(a%d)", offset2, reg2);
                move_mem(offset1, reg1, part1);
                return 0;
            }    
        }
        if (ptype1 == PT_MEM) {
            if (ptype2 == PT_BYTEREG) {
                load_addr(offset1, 1, addr);
                get_bytereg(reg2, part2, where);
                fprintf(outfile, "        move.b  %s,%s\n", where, addr);
                return 0;
            }
            if (ptype2 == PT_WORDREG) {
                load_addr(offset1, 1, addr);
                fprintf(outfile, "        move.w  d%d,d0\n", reg2);
                fprintf(outfile, "        lsr.w   #8,d0\n");
                fprintf(outfile, "        move.b  d%d,%s\n", reg2, addr);
                fprintf(outfile, "        move.b  d0,1+%s\n", addr);
                return 0;
            }
        }

        if (ptype1 == PT_WORDREG) {
            if (ptype2 == PT_CONST) {
                x = 0;
                if (smc2) {
                    expr = eval_expression(offset2);
                    x = smc_offset(expr);
                }
                fprintf(outfile, "        move.w  #%s+%d,d%d\n", offset2, x, reg1);
                if (current->addrset[reg1] == AC_SPEC) {
                    load_addr(offset2, reg1, addr);
                    fprintf(outfile, "        lea     %s,a%d\n", addr, reg1);
                }
                return 0;
            }
            if (ptype2 == PT_MEM) {
                load_addr(offset2, 1, addr);
                fprintf(outfile, "        move.b  1+%s,d%d\n", addr, reg1);
                fprintf(outfile, "        asl.w   #8,d%d\n", reg1);
                fprintf(outfile, "        move.b  %s,d%d\n", addr, reg1);
                return 0;
            }
        }
    }

    /************** INC, DEC */

    if ((eq(txt, "inc")) || (eq(txt, "dec"))) {
        if (eq(txt, "dec"))
            strcpy(cmd, "sub");
        else
            strcpy(cmd, "add");
        if (ptype1 == PT_BYTEREG) {
            if (part1 == RP_LOW)
                fprintf(outfile, "        %sq.b  #1,d%d\n", cmd, reg1);
            else {
                fprintf(outfile, "        %s.w   #$100,d%d\n", cmd, reg1);
                if (current->do_save_flags) {
                    fprintf(outfile, "        move.w  d%d,d1\n", reg1);
                    fprintf(outfile, "        and.w   #$ff00,d1\n");
                }
            }
            VFLAGSAVE
            return 0;
        }
        if (ptype1 == PT_WORDREG) {
            fprintf(outfile, "        %sq.w  #1,d%d\n", cmd, reg1);
            if (current->addrset[reg1] == AC_SPEC)
                fprintf(outfile, "        %sq.w  #1,a%d\n", cmd, reg1);
            return 0;
        }
        if (ptype1 == PT_DEREF) {
            LEA(reg1)
            fprintf(outfile, "        %sq.b  #1,%s(a%d)\n", cmd, offset1, reg1);
            VFLAGSAVE
            return 0;
        }

        if (ptype1 == PT_SP) {
            if (eq(txt, "dec"))
               fprintf(outfile, "        __DECSTACK\n");
            else
               fprintf(outfile, "        __INCSTACK\n");
            return 0;
        }
    }

    /************** RRC, RLC, SRL, SLA, SLA, SRA, RR, RL, SET, RES */

    if ((eq(txt, "rr")) || (eq(txt, "rl")) || (eq(txt, "rrc")) ||
       (eq(txt, "rlc")) || (eq(txt, "srl")) || (eq(txt, "sla")) ||
       (eq(txt, "sra"))) {

        if (ptype1 == PT_BYTEREG) {
            reg2 = reg1;
            if (part1 == RP_HIGH) {
                fprintf(outfile, "        move.w  d%d,(a1)\n", reg1);
                fprintf(outfile, "        move.b  (a1),d1\n");
                reg2 = 1;
            }
            if (eq(txt, "rr")) {
                fprintf(outfile, "        move.b  (a0),ccr\n");
                fprintf(outfile, "        roxr.b  #1,d%d\n", reg2);
            }
            if (eq(txt, "rl")) {
                fprintf(outfile, "        move.b  (a0),ccr\n");
                fprintf(outfile, "        roxl.b  #1,d%d\n", reg2);
            }
            if (eq(txt, "rrc"))
                fprintf(outfile, "        ror.b   #1,d%d\n", reg2);
            if (eq(txt, "rlc"))
                fprintf(outfile, "        rol.b   #1,d%d\n", reg2);
            if (eq(txt, "sra"))
                fprintf(outfile, "        asr.b   #1,d%d\n", reg2);
            if (eq(txt, "sla"))
                fprintf(outfile, "        add.b   d%d,d%d\n", reg2, reg2);
            if (eq(txt, "srl"))
                fprintf(outfile, "        lsr.b   #1,d%d\n", reg2);

            VFLAGSAVE

            if (part1 == RP_HIGH) {
                fprintf(outfile, "        move.b  d1,(a1)\n");
                fprintf(outfile, "        move.w  (a1),d%d\n", reg1);
            }

            return 0;                           
        }
    }

    /************** AND, OR, XOR, SUB, CP */

    if ((eq(txt, "and")) || (eq(txt, "or")) || (eq(txt, "xor")) || (eq(txt, "sub")) || (eq(txt, "cp"))) {
        strcpy(cmd, txt);
        if (eq(txt, "cp"))
            strcpy(cmd, "cmp");
        if (eq(txt, "xor"))
            strcpy(cmd, "eor");
        if (ptype1 == PT_BYTEREG) {
            get_bytereg(reg1, part1, where);
            if ((where[0] == '(') && (eq(txt, "xor"))) {
                fprintf(outfile, "        move.b  %s,d1\n", where);
                fprintf(outfile, "        eor.b   d1,d7\n");
            } else
                fprintf(outfile, "        %s.b   %s,d7\n", cmd, where);
            VFLAGSAVE
            return 0;
        }
        if (ptype1 == PT_CONST) {
            fprintf(outfile, "        %s.b   #%s,d7\n", cmd, offset1);
            VFLAGSAVE
            return 0;
        }
        if (ptype1 == PT_DEREF) {
            LEA(reg1)
            if (eq(txt, "xor")) {
                fprintf(outfile, "        move.b  %s(a%d),d1\n", offset1, reg1);
                fprintf(outfile, "        eor.b   d1,d7\n");
            } else
                fprintf(outfile, "        %s.b   %s(a%d),d7\n", cmd, offset1, reg1);
            VFLAGSAVE
            return 0;
        }
    }

    /************** BIT, SET, RES */

    if ((eq(txt, "bit")) || (eq(txt, "set")) || (eq(txt, "res"))) {
        strcpy(cmd, "btst");
        if (eq(txt, "set"))
            strcpy(cmd, "bset");
        if (eq(txt, "res"))
            strcpy(cmd, "bclr");
        if (ptype2 == PT_BYTEREG) {
            if (part1 == RP_HIGH)
                fprintf(outfile, "        %s    #8+%s,d%d\n", cmd, a1, reg2);
            else
                fprintf(outfile, "        %s    #%s,d%d\n", cmd, a1, reg2);
            if (eq(txt, "bit"))
                VFLAGSAVE
            return 0;
        }
        if (ptype2 == PT_DEREF) {
            LEA(reg2)
            fprintf(outfile, "        %s    #%s,%s(a%d)\n", cmd, a1, offset2, reg2);
            if (eq(txt, "bit"))
                VFLAGSAVE
            return 0;
        }
    }

    /************** CALL, JP, JR */

    if ((eq(txt, "jp")) || (eq(txt, "jr")) || (eq(txt, "call"))) {
        strcpy(cmd, "jmp  ");
        if (eq(txt, "call"))
            strcpy(cmd, "jsr  ");

        if (ptype1 == PT_DEREF) {
            LEA(reg1)
            fprintf(outfile, "       jmp     %s(a%d)\n", offset1, reg1);
            return 0;
        }

        strcpy(cmd2, empty);
        strcpy(dest, offset1);
        decode_cond(a1, cmd2);

        if (*cmd2) {
            strcpy(dest, offset2);
            x = symbol_check(dest);

            y = 0;
            if ((exact_stack) && (x == ST_LABEL) && (eq(cmd, "jsr  ")))
                strcpy(cmd, "__JSR");
            if ((exact_stack) && (x == ST_ROMCALL) && (eq(cmd, "jmp  "))) {
                strcpy(cmd, "jsr  ");
                y = 1;
            }

            if (!(no_flags))
                if (!(current->gets_flags))
                    fprintf(outfile, "        move.b  (a0),ccr\n");

            if ((x == ST_LABEL) && (eq(cmd, "jmp  "))) {
                decode_cond_direct(a1, cmd);
                fprintf(outfile, "        %s   %s\n", cmd, dest);
            } else {
                fprintf(outfile, "        %s     NC__%d\n", cmd2, ncc);
                fprintf(outfile, "        %s   %s\n", cmd, dest);
                fprintf(outfile, "NC__%d:\n", ncc++);
            }

            if (y)
                fprintf(outfile, "        __RTS\n");
        } else {
            x = symbol_check(dest);

            y = 0;
            if (x == ST_LABEL) {
                if (eq(cmd, "jsr  ")) {
                    if (exact_stack) {
                        strcpy(cmd, "__JSR");
                    } else {
                        strcpy(cmd, "bsr  ");
                    }
                } else {
                    strcpy(cmd, "bra  ");
                }
            }

            if ((exact_stack) && (x == ST_ROMCALL) && (eq(cmd, "jmp  "))) {
                strcpy(cmd, "jsr  ");
                y = 1;
            }
                     
            fprintf(outfile, "        %s   %s\n", cmd, dest);

            if (y)
                fprintf(outfile, "        __RTS\n");
        }
        return 0;
    }

    /************** DJNZ */

    if (eq(txt, "djnz")) {
        fprintf(outfile, "        sub.w   #$100,d2\n");
        fprintf(outfile, "        move.w  d2,d1\n");
        fprintf(outfile, "        and.w   #$ff00,d1\n");
        fprintf(outfile, "        bne     RA_%s\n", a1);
        return 0;
    }

    return 1;
}

/****************** Main code converter ***********************************/

void convert_code()
{
    source_item* current;
    current = z80_source;
    while (current) {
        switch (current->type) {

            case ITEM_LINE:
                if (show_original)
                    fprintf(outfile, "**** Original Line: %s\n", txt);
                break;

            case ITEM_LABEL:
                if (zkart3dhack) {
                    if (eq(txt, "title_message"))
                        zkart3dhack = 3;
                    if (eq(txt, "clab1")) {
                        fprintf(outfile, "        move.b  RA_clab1(pc),d0\n");
                        fprintf(outfile, "        cmp.b   #$ed,d0\n");
                        fprintf(outfile, "        bne.s   no_negate\n");
                        fprintf(outfile, "        neg.b   d7\n");
                        fprintf(outfile, "no_negate:\n");
                        if (exact_stack)
                            fprintf(outfile, "        __RTS\n");
                        else
                            fprintf(outfile, "        rts\n");
                    }
                }
                fprintf(outfile, "        __LABEL %s\n", txt);
                break;

            case ITEM_EQUATE:
                fprintf(outfile, "%s    equ     %s\n", txt, a1);
                break;

            case ITEM_INSTRUCTION:
                if (zkart3dhack == 1) {
                    fprintf(outfile, "        bra.s   RA_title_screen\n");
                    zkart3dhack = 2;
                }
                if (zkart3dhack == 2)
                    break;
                if (convert_instruction(current)) {
                    fprintf(outfile, "******** WARNING: Error on instruction %s %s, %s\n", txt, a1, a2);
                    printf("******** WARNING: Error on instruction %s %s, %s\n", txt, a1, a2);
                }
                break;

            case ITEM_DIRECTIVE:
                if (eq(txt,"end")) {
                } else if (eq(txt, "org")) {
                } else if (eq(txt, "define")) {
                } else {
                    fprintf(outfile, "******** WARNING: Unknown directive %s\n", txt);
                    printf("******** WARNING: Unknown directive %s\n", txt);
                }    
                break;

            case ITEM_CONSTNUM:
                fprintf(outfile, "        dc.b    %s\n", txt);
                break;

            case ITEM_CONSTSTR:
                fprintf(outfile, "        dc.b    \"%s\"\n", txt);
                break;

            case ITEM_MACROCALL:
                fprintf(outfile, "******** WARNING: Unhandled macro %s\n", txt);
                printf("******** WARNING: Unhandled macro %s\n", txt);
                break;

            case ITEM_CONSTWORD:
                fprintf(outfile, "        __DW    %s\n", txt);
                break;
        }
        current = current->next;
    }
}

/****************** Build beginning / end of file ************************/

void output_68k()
{
    fprintf(outfile, "**** This program converted from Z80 to 68K by Source Guru 0.6\n");
    if (exact_stack)
        fprintf(outfile, "EXACT_STACK equ 1\n");
    if (turbo)
        fprintf(outfile, "        INCLUDE \"sim-ti85.h\"\n");
    else
        fprintf(outfile, "        INCLUDE \"sim_ti85.h\"\n");
    convert_code();
    if (exact_stack)
        if (turbo)
            fprintf(outfile, "        EVEN\n___byebye:\n        jsr     ti85lib3@0012\n");
        else
            fprintf(outfile, "        EVEN\n___byebye:\n        jsr     ti85lib2@0012\n");
    fprintf(outfile, "program_end:\n");
    if (turbo) {
        fprintf(outfile, "        SECTION bss,BSS\n");
        fprintf(outfile, "buffer: ds.b    $8100\n");
    }
    fprintf(outfile, "        END\n");
}
