#include <stdio.h>

#include "srcguru.h"

#include "hash.h"

#include "eval.h"

#include "params.h"

#include "symbols.h"

/****************** Check if symbol is a ROM_CALL ***********************/

int symbol_check(char* sym)
{
    hash_item* x;
    x = get_symbol(sym);
    if (x) {
        strcpy(sym, x->value);
        return x->type;
    }
    return 0;
}

/***************** Parameter decoding **********************************/

typedef struct {
    int ptype;
    int reg;
    int part;
    char name[4];
} regdata;

regdata regs[18] = {{PT_BYTEREG, 7, RP_LOW, "a"},
                    {PT_BYTEREG, 2, RP_HIGH, "b"},
                    {PT_BYTEREG, 2, RP_LOW, "c"},
                    {PT_BYTEREG, 3, RP_HIGH, "d"},
                    {PT_BYTEREG, 3, RP_LOW, "e"},
                    {PT_BYTEREG, 4, RP_HIGH, "h"},
                    {PT_BYTEREG, 4, RP_LOW, "l"},
                    {PT_BYTEREG, 5, RP_HIGH, "ixh"},
                    {PT_BYTEREG, 5, RP_LOW, "ixl"},
                    {PT_BYTEREG, 6, RP_HIGH, "iyh"},
                    {PT_BYTEREG, 6, RP_LOW, "iyl"},
                    {PT_WORDREG, 2, RP_ALL, "bc"},
                    {PT_WORDREG, 3, RP_ALL, "de"},
                    {PT_WORDREG, 4, RP_ALL, "hl"},
                    {PT_WORDREG, 5, RP_ALL, "ix"},
                    {PT_WORDREG, 6, RP_ALL, "iy"},
                    {PT_AF, -1, RP_ALL, "af"},
                    {PT_SP, -1, RP_ALL, "sp"}};

/***************** Delete a character from a string ***********************/

void delete_char(char* x)
{
    while (*x) {
        *x = *(x + 1);
        x++;
    }
}

/***************** Remove blanks from text ********************************/

void remove_blanks(char* s)
{
    while (*s) 
        if ((*s == ' ') || (*s == '\t'))
            delete_char(s);
        else
            s++;
}

/***************** Remove relocation ampersands from text *****************/

void remove_ampersands(char* data)
{
    while (*data == '&')
        delete_char(data);
        
    if (*data++ == 0)
        return;

    while (*data)
        if ((*data == '&') && (*(data - 1) == '('))
            delete_char(data);
        else
            data++;
}

/****************** Decode a parameter to an instruction ******************/

void decode_param(char* dt, int* ptype, int* reg, int* part, char* offset)
{
    char data[MAX_LINE];
    int x, y = -5;
    *reg = -1;
    *part = -1;
    *offset = 0;

    strcpy(data, dt);
    remove_blanks(data);
    remove_ampersands(data);

    if (data[0] == 0) {
        *ptype = PT_NULL;
        return;
    }

    if (data[0] == '(') {
        y = 0;
        for (x = 0; x < strlen(data); x++) {
            if (data[x] == '(') y++;
            if ((x != (strlen(data) - 1)) && (data[x] == ')')) y--;
            if (y == 0) break;
        }
    }
        
    if (y != 1) {
        *ptype = PT_CONST;
        /* if (data[0] == '&') {
            strcpy(offset, data + 1);
            return;
        } */

        for (x = 0; x < 18; x++)
            if (eq(data, regs[x].name)) {
                *ptype = regs[x].ptype;
                *reg = regs[x].reg;
                *part = regs[x].part;
                return;
            }
        strcpy(offset, data);
        y = 0;
        for (x = 0; x < strlen(offset); x++) {
            if (offset[x] == '(') y++;
            if (offset[x] == ')') y--;
            if (y == -1) {
                offset[x] = 0;
                break;
            }    
        }
    } else {
        *ptype = PT_MEM;
        data[strlen(data) - 1] = 0;
        strcpy(offset, data + 1);

        if (offset[0] == '&') {
            strcpy(offset, data + 2);
            return;
        }

        if (eq(offset, "bc")) {
            *ptype = PT_DEREF;
            *reg = 2;
        }
        if (eq(offset, "de")) {
            *ptype = PT_DEREF;
            *reg = 3;
        }
        if (eq(offset, "hl")) {
            *ptype = PT_DEREF;
            *reg = 4;
        }
        if (eq(offset, "ix")) {
            *ptype = PT_DEREF;
            *reg = 5;
        }
        if (eq(offset, "iy")) {
            *ptype = PT_DEREF;
            *reg = 6;
        }
        if (*ptype == PT_DEREF)
            strcpy(offset, "");
        if (!(strnicmp(offset, "ix+", 3))) {
            *ptype = PT_DEREF;
            *reg = 5;
            strcpy(offset, data + 4);
        }
        if (!(strnicmp(offset, "ix-", 3))) {
            *ptype = PT_DEREF;
            *reg = 5;
            strcpy(offset, data + 3);
        }
        if (!(strnicmp(offset, "iy+", 3))) {
            *ptype = PT_DEREF;
            *reg = 6;
            strcpy(offset, data + 4);
        }
        if (!(strnicmp(offset, "iy-", 3))) {
            *ptype = PT_DEREF;
            *reg = 6;
            strcpy(offset, data + 3);
        }
    }            
}

/****************** Access needed data ***********************************/

void get_bytereg(int reg, int part, char *where)
{
    if (part == RP_HIGH) {
        fprintf(outfile, "        move.w  d%d,(a1)\n", reg);
        strcpy(where, "(a1)");
    } else {
        *where++ = 'd';
        *where++ = '0' + reg;
        *where++ = 0;
    }
}

void put_bytereg(int src, int dest, int destpart)
{
    if (destpart == RP_LOW) {
        if (src != dest)
            fprintf(outfile, "        move.b  d%d,d%d\n", src, dest);
    } else {
        if (src > 1) {
            fprintf(outfile, "        move.b d%d,d1\n", src);    
            src = 1;
        }
        fprintf(outfile, "        asl.w   #8,d%d\n", src);
        fprintf(outfile, "        and.w   #$ff,d%d\n", dest);
        fprintf(outfile, "        or.w    d%d,d%d\n", src, dest);
    }
}

void move_mem(char *mem, int reg, int part)
{
    if (part == RP_HIGH) {
        fprintf(outfile, "        move.b  %s,d1\n", mem);
        put_bytereg(1, reg, part);
    } else {
        fprintf(outfile, "        move.b  %s,d%d\n", mem, reg);
    }
}

/****************** Load a precalculated address into A1 ******************/

void load_addr(char *addr, int reg, char *ea)
{
    char s[MAX_LINE];
    ev* addrv;
    addrv = eval_expression(addr);
    exp_to_str(addrv, s);

    switch (classify_expression(addrv)) {
        case ET_LABEL:
            if (turbo)
                sprintf(ea, "%s+%d-$10000(a0)", addr, smc1 * smc_offset(addrv));
            else
                sprintf(ea, "_comment-$9000+%s+%d", addr, smc1 * smc_offset(addrv));
            break;
        case ET_ERROR:
            fprintf(outfile, "******** WARNING: Confusing expression %s\n", addr);
            printf("******** WARNING: Confusing expression %s\n", addr);
        default:
            if (turbo)
                sprintf(ea, "%s-$10000(a0)", addr);
            else
                sprintf(ea, "((%s-$7000)&$7FFF)(a0)", addr);
            break;
    }
}

/****************** Check for self-modifying code ***********************/

int smc_offset(ev* expr)
{
    source_item* x;
    hash_item* l;

    int ptype;
    int reg;
    int part;
    char offset[MAX_LINE];

    if (classify_expression(expr) != ET_LABEL)
        return 0;
    if (expr->first_label->next_label != NULL)
        return 0;
    if (expr->num != 1)
        return 0;
    l = get_symbol(expr->first_label->name);
    x = (source_item *) l->misc;

    while (x) {
        if (x->type == ITEM_CONSTNUM)
            break;
        if (x->type == ITEM_CONSTSTR)
            break;
        if (x->type == ITEM_CONSTWORD)
            break;
        if (x->type == ITEM_INSTRUCTION) {
            return 2;
        }
        x = x->next;
    }
    return 0;
}

/****************** Decode (and invert) a conditional *********************/

void decode_cond(char* in, char* out)
{
    if (eq(in, "z"))
        strcpy(out, "bne");
    if (eq(in, "nz"))
        strcpy(out, "beq");
    if (eq(in, "c"))
        strcpy(out, "bcc");
    if (eq(in, "nc"))
        strcpy(out, "bcs");
    if (eq(in, "p"))
        strcpy(out, "blt");
    if (eq(in, "m"))
        strcpy(out, "bge");
}

/****************** Decode (and invert) a conditional *********************/

void decode_cond_direct(char* in, char* out)
{
    if (eq(in, "z"))
        strcpy(out, "beq");
    if (eq(in, "nz"))
        strcpy(out, "bne");
    if (eq(in, "c"))
        strcpy(out, "bcs");
    if (eq(in, "nc"))
        strcpy(out, "bcc");
    if (eq(in, "p"))
        strcpy(out, "bge");
    if (eq(in, "m"))
        strcpy(out, "blt");
}

