#include "srcguru.h"

#include "parser.h"

char empty[] = "";
int program_mode = MODE_NORMAL;
int show_original = 1;
int exact_stack = 0;
int no_flags = 0;
int smc1 = 0;
int smc2 = 0;
int smc3 = 0;
int turbo = 0;
int zkart3dhack = 0;
source_item *z80_source = NULL;
source_item *last_item = NULL;

int stricmp(char *s1, char *s2)
{
    while ((*s1) || (*s2)) {
        if (tolower(*s1) > tolower(*s2))
            return 1;
        if (tolower(*s1++) < tolower(*s2++))
            return -1;
    }
    return 0;
}

int strnicmp(char *s1, char *s2, int n)
{
    while (((*s1) || (*s2)) && (n--)) {
        if (tolower(*s1) > tolower(*s2))
            return 1;
        if (tolower(*s1++) < tolower(*s2++))
            return -1;
    }
    return 0;
}

void delete_item(source_item *item)
{
    source_item *prev;
    source_item *next;

    prev = item->prev;
    next = item->next;

    free(item->maintext);
    free(item->arg1);
    free(item->arg2);
    free(item);

    if (prev == NULL)
        z80_source = next;
    else
        prev->next = next;

    if (next == NULL)
        last_item = prev;
    else
        next->prev = prev;
}

source_item* _add_item(int type, char *maintext, char *arg1, char *arg2, source_item *prev, source_item *next)
{
    source_item* new_item = NULL;

    new_item = (source_item *) malloc(sizeof(source_item));

    if (prev == NULL) 
        z80_source = new_item;
    else
        prev->next = new_item;

    if (next == NULL)
        last_item = new_item;
    else
        next->prev = new_item;

    new_item->maintext = (char *) malloc(strlen(maintext) + 1);
    strcpy(new_item->maintext, maintext);
    new_item->arg1 = (char *) malloc(strlen(arg1) + 1);
    strcpy(new_item->arg1, arg1);
    new_item->arg2 = (char *) malloc(strlen(arg2) + 1);
    strcpy(new_item->arg2, arg2);

    new_item->prev = prev;
    new_item->next = next;
    new_item->type = type;

    return new_item;
}

void add_item(int type, char *maintext, char *arg1, char *arg2)
{
    _add_item(type, maintext, arg1, arg2, last_item, NULL); 
}

int main(int argc, char *argv[])
{
    int argnum;

    printf("Source Guru version 0.6 - Z-80 assembly Source converter\n\n");
    printf("Copyright 1999-2000 by Patrick Davidson\n\n");
    
    if (argc < 3) {
        printf("Usage: %s infile outfile [options]\n\n", argv[0]);
        printf("Main program options\n");
        printf("debug          Write parser debugging info instead of converting\n");
        printf("cleanup        Make cleaned up Z80 source instead of converting\n");
        printf("nooriginal     Don't put original lines as comments in 68K output\n");
        printf("\nCode conversion options - Refer to manual for complete descriptions\n");
        printf("exs            Stores stack return values as Z80 addresses\n");
        printf("smc1           Handle self-modifying code like 'ld (&code+1),a'\n");
        printf("smc2           Handle self-modifying code like 'ld hl,&code+1 \\ ld (hl), a'\n");
        printf("turbo          Makes programs faster but require 32K free memory\n");
        printf("\nProgram-specifc hacks - Refer to manual for complete descriptions\n");
      /*  printf("zkart3dhack    Compatibility hack required for Z-Kart 3D\n");  */
        return 666;
    } else {
        if (infile = fopen(argv[1], "r")) {
            if (outfile = fopen(argv[2], "w")) {
                argnum = 3;

                while (argnum < argc) {
                    if (eq(argv[argnum], "debug"))
                        program_mode = MODE_DEBUG;
                    if (eq(argv[argnum], "cleanup"))
                        program_mode = MODE_CLEANUP;
                    if (eq(argv[argnum], "nooriginal"))
                        show_original = 0;
                    if (eq(argv[argnum], "exs"))
                        exact_stack = 1;
                    if (eq(argv[argnum], "smc1"))
                        smc1 = 1;
                    if (eq(argv[argnum], "smc2"))
                        smc2 = 1;
                    if (eq(argv[argnum], "turbo"))
                        turbo = 1;
                    if (eq(argv[argnum], "zkart3dhack")) {
                       printf("******** WARNING : This hack is not yet completely stable !!!\n");
                        zkart3dhack = 1;
                        smc1 = 1;
                        exact_stack = 1;
                    }
                    argnum++;
                }

                if (parse_file()) {
                    switch (program_mode) {
                        case MODE_NORMAL:
                            do_vflags();
                            fix_align();
                            addr_optimize();
                            output_68k();   
                            break;           
                        case MODE_CLEANUP:
                            output_z80();
                            break;
                        case MODE_DEBUG:
                            output_debug();
                            break;
                    }
                }

            } else {
                printf("Unable to open output file!\n");
            }
            fclose(infile);
        } else {
            printf("Unable to open input file!\n");
        }
    }
}
