#include "srcguru.h"

#include "eval.h"

#include "hash.h"

#include "symbols.h"

#include "params.h"

#define FS_CHANGES 1
#define FS_USES 2
#define FS_NOTHING 3

int flag_status(source_item *current, int which)
{
    if (current == NULL)
        return FS_NOTHING;
    if (current->type == ITEM_INSTRUCTION) {
        if (current->uses_flags)
            return FS_USES;
        if ((current->sets_flags == SF_ALL) || (current->sets_flags == which))
            return FS_CHANGES;
    }
    return flag_status(current->next, which);
}

void do_vflags()
{
    int type, reg, part;
    char offset[MAX_LINE];

    source_item *current;
    current = z80_source;

    /* Classify instructions according to flag use */

    while (current) {
        current->sets_flags = SF_NONE;
        current->leaves_flags = 0;
        current->uses_flags = 0;
        current->do_save_flags = 0;
        current->gets_flags = 0;

        if ((!(no_flags)) && (current->type == ITEM_INSTRUCTION)) {

            if ((eq(txt, "sub")) ||
                (eq(txt, "or")) ||
                (eq(txt, "and")) ||
                (eq(txt, "cp")) ||
                (eq(txt, "xor"))) {
                current->sets_flags = SF_ALL;
                current->leaves_flags = 1;
            }

            if ((eq(txt, "neg")) || (eq(txt, "cpl"))) {
                current->sets_flags = SF_ALL;
                current->leaves_flags = 1;
            }

            if ((eq(txt, "add")) || (eq(txt, "inc")) || (eq(txt, "dec"))) {
                decode_param(a1, &type, &reg, &part, offset);
                if (type != PT_WORDREG) {
                    current->sets_flags = SF_ALL;
                    if (!((type == PT_BYTEREG) && (part == RP_HIGH)))
                        current->leaves_flags = 1;
                }
            }

            if ((eq(txt, "rlc")) ||
                (eq(txt, "rrc")) ||
                (eq(txt, "sla")) ||
                (eq(txt, "sra")) ||
                (eq(txt, "srl"))) {
                decode_param(a1, &type, &reg, &part, offset);
                current->sets_flags = SF_ALL;
                if (!((type == PT_BYTEREG) && (part == RP_HIGH)))
                    current->leaves_flags = 1;
            }
               
            if ((eq(txt, "rl")) || (eq(txt, "rr"))) {
                decode_param(a1, &type, &reg, &part, offset);
                current->sets_flags = SF_ALL;
                current->uses_flags = 1;
                if (!((type == PT_BYTEREG) && (part == RP_HIGH)))
                    current->leaves_flags = 1;
            }

            if (eq(txt, "ccf")) {
                current->uses_flags = 1;
                current->sets_flags = SF_CARRY;
            }

            if ((eq(txt, "rlca")) || (eq(txt, "rrca"))) 
                current->sets_flags = SF_ALL;

            if (eq(txt, "scf"))
                current->sets_flags = SF_CARRY;

            if ((eq(txt, "jr")) ||
                (eq(txt, "jp")) ||
                (eq(txt, "call")) ||
                (eq(txt, "ret"))) {
                strcpy(offset, "");
                decode_cond(a1, offset);
                if (*offset) {
                    current->uses_flags = 1;
                } else {
                    current->uses_flags = -1;
                    if (!(eq(txt, "ret"))) {
                        decode_param(a1, &type, &reg, &part, offset);
                        if (symbol_check(offset) == ST_ROMCALL) {
                            current->uses_flags = 0;
                            current->sets_flags = SF_ALL;
                        }
                    }
                }
            }

            if (eq(txt, "bit"))
                current->sets_flags = SF_ALL;
   
        }
        current = current->next;
    }

    if (no_flags)
        return;

    /* Decide which instructions really must save flag changes */

    current = z80_source;
    while (current) {
        if (current->type == ITEM_INSTRUCTION) {
            current->do_save_flags = (flag_status(current->next, current->sets_flags) != FS_CHANGES);
        }
        current = current->next;
    }

    /* Decide which flag-using instructions can use pre-loaded flags */

    current = z80_source;
    while (current) {
        if (current->type == ITEM_INSTRUCTION) {
            if (current->leaves_flags) {
                while ((current) && (current->type != ITEM_LABEL)) {
                    current = current->next;
                    if (current->type == ITEM_INSTRUCTION) {
                        current->gets_flags = 1;
                        break;
                    }
                }
            } else
                current = current->next;
        } else
            current = current->next;
    }
}

